# Python Modules
import logging
import os

# 3rd Party Modules
import psutil

# Project Modules
from sleepguard.monitors import debug
from sleepguard.monitors.base import UtilizationMonitor


log = logging.getLogger(__name__)


class CpuLoadMonitor(UtilizationMonitor):
    """
    Keeps track of the average CPU load over the given ``threshold_period``.
    """

    _index_lookup = {
        1: 0,
        5: 1,
        15: 2
    }

    def __init__(
            self,
            poll_period: float,
            log_history_size: int,
            threshold: float,
            *args,
            **kwargs
    ):
        # Note the poll period is ignored.
        if log_history_size not in (1, 5, 15):
            raise ValueError(
                f"The CPU load average threshold period must be 1 or 5 or 15. "
                f"Got: {log_history_size}"
            )

        super().__init__(poll_period, log_history_size, threshold, *args, **kwargs)

        self._load_index = self._index_lookup[int(log_history_size)]

    def run(self):
        # Just exit the loop, since we aren't the ones actually monitoring
        # things.
        self.stop()

    @debug
    def get_instantaneous_value(self) -> float:
        """
        Returns the current load average reported by the OS.

        :return:
        """
        return os.getloadavg()[self._load_index]

    def get_average_utilization(self) -> float:
        """
        Get the load average over the ``threshold_period``.

        :return:
        """
        return self.get_instantaneous_value()


class CpuUtilizationMonitor(UtilizationMonitor):
    """
    Keeps track of the average CPU percent utilization over the given
    ``threshold_period``.
    """
    def __init__(
            self,
            poll_period: float,
            log_history_size: int,
            threshold: float,
            *args,
            **kwargs
    ):
        super().__init__(poll_period, log_history_size, threshold, *args, **kwargs)

    @debug
    def get_instantaneous_value(self) -> float:
        return max(psutil.cpu_percent(percpu=True))



# Python Modules
import logging
import os
import time

# 3rd Party Modules
import psutil

from sleepguard.monitors import debug
# Project Modules
from sleepguard.monitors.base import UtilizationMonitor


log = logging.getLogger(__name__)


class NetUtilizationMonitor(UtilizationMonitor):
    """
    Keeps track of the average total number of bytes sent and received per
    second over the given ``threshold_period``.
    """
    def __init__(
            self,
            poll_period: float,
            log_history_size: int,
            threshold: float,
            *args,
            **kwargs
    ):
        super().__init__(poll_period, log_history_size, threshold, *args, **kwargs)

        # psutil.net_io will return the total number of bytes since the system
        # started. To get a snapshot we need to calculate the difference
        # between the current and previous value, which is initialized here
        # at the start of the monitor.
        self._previous_value = self._get_total_bytes()

    def run(self):
        # We're going to sleep here before starting the loop otherwise the
        # first instantaneous value reading will likely be 0.
        time.sleep(self.poll_period)

        super().run()

    @debug
    def get_instantaneous_value(self) -> float:
        total_bytes = self._get_total_bytes()
        bytes_per_second = (total_bytes - self._previous_value) / self._poll_period

        with self._lock:
            self._previous_value = total_bytes

        return bytes_per_second

    @staticmethod
    def _get_total_bytes() -> int:
        net_io = psutil.net_io_counters()
        if net_io is None:
            return 0

        total_bytes = net_io.bytes_sent + net_io.bytes_recv

        return total_bytes



# Python Modules
import logging

from sleepguard.monitors import DEFAULT_CPU_LOAD_LOG_HISTORY_SIZE, DEFAULT_CPU_LOAD_THRESHOLD, \
    DEFAULT_CPU_UTIL_THRESHOLD, \
    DEFAULT_GPU_UTIL_THRESHOLD, DEFAULT_NET_UTIL_THRESHOLD, MonitorCriteria

# 3rd Party Modules

# Project Modules


DEFAULT_LOG_LEVEL = logging.INFO
DEFAULT_LOG_FORMAT = "%(asctime)-15s [%(name)s]:%(lineno)d %(levelname)s %(message)s"
DEFAULT_MONITOR_CRITERIA: list[MonitorCriteria] = [
    MonitorCriteria(
        system_resource="cpu_load",
        threshold=DEFAULT_CPU_LOAD_THRESHOLD,
        log_history_size=DEFAULT_CPU_LOAD_LOG_HISTORY_SIZE,
    ),
    MonitorCriteria(system_resource="cpu_util", threshold=DEFAULT_CPU_UTIL_THRESHOLD),
    MonitorCriteria(system_resource="gpu_util", threshold=DEFAULT_GPU_UTIL_THRESHOLD),
    MonitorCriteria(system_resource="net_util", threshold=DEFAULT_NET_UTIL_THRESHOLD)
]



# Python Modules
import logging
from typing import Optional

# 3rd Party Modules

from sleepguard.monitors import MonitorCriteria
from sleepguard.monitors.base import UtilizationMonitor
from sleepguard.monitors.cpu import CpuLoadMonitor, CpuUtilizationMonitor
from sleepguard.monitors.gpu import GpuUtilizationMonitor
from sleepguard.monitors.net import NetUtilizationMonitor

# Project Modules


log = logging.getLogger(__name__)


class UtilizationMonitorFactory:
    """
    A factory class for creating :class:`~monitors.base.UtilizationMonitor`
    instances.
    """

    @staticmethod
    def from_criteria(
            criteria: MonitorCriteria,
            raise_on_error: bool = False
    ) -> Optional[UtilizationMonitor]:
        """
        Create an instance of a :class:`~monitors.base.UtilizationMonitor`
        from the given ``criteria``.

        :param criteria: The ``MonitorCriteria`` to used to create the instance.
        :param raise_on_error: If ``True``, an exception will be raised if
               the instance cannot be created, for example an unsupported
               GPU is detected when using the ``GpuUtilizationMonitor``.
               Otherwise, ``None`` is returned.
        :return:
        """
        if criteria.system_resource == "cpu_load":
            monitor_cls = CpuLoadMonitor
        elif criteria.system_resource == "cpu_util":
            monitor_cls = CpuUtilizationMonitor
        elif criteria.system_resource == "gpu_util":
            has_gpu = len(GpuUtilizationMonitor.get_device_types()) > 0
            if has_gpu or raise_on_error:
                monitor_cls = GpuUtilizationMonitor
            else:
                return None
        elif criteria.system_resource == "net_util":
            monitor_cls = NetUtilizationMonitor
        else:
            raise ValueError("Unsupported system resource type")

        return monitor_cls(**criteria.as_kwargs())



# Python Modules
import logging

from typing import Literal

from sleepguard.monitors import debug
# 3rd Party Modules

# Project Modules
from sleepguard.monitors.base import UtilizationMonitor


log = logging.getLogger(__name__)


GpuType = Literal["amd", "nvidia"]


class GpuUtilizationMonitor(UtilizationMonitor):
    """
    Keeps track of the average GPU percent utilization over the given
    ``threshold_period``.
    """
    def __init__(
            self,
            poll_period: float,
            log_history_size: int,
            threshold: float,
            *args,
            **kwargs
    ):
        super().__init__(poll_period, log_history_size, threshold, *args, **kwargs)

        self.device_types = self.get_device_types()

        if not self.device_types:
            raise ValueError("No GPU devices found.")
        else:
            log.debug(f"Found GPU devices: {self.device_types}")

    @staticmethod
    def get_device_types() -> list[GpuType]:
        """
        Get the list of available GPU types installed on the system.

        :return:
        """
        devices: list[GpuType] = []
        try:
            import GPUtil

            if GPUtil.getGPUs():
                devices.append("nvidia")
        except Exception as e:
            log.debug(f"Unable to find any NVIDIA GPUs: {e}")

        try:
            import pyadl
            if pyadl.ADLManager.getInstance().getDevices():
                devices.append("amd")
        except Exception as e:
            log.debug(f"Unable to find any AMD GPUs: {e}")

        return devices

    @debug
    def get_instantaneous_value(self) -> float:
        """
        Returns the max utilization over all the detected GPU devices.

        :return:
        """
        values = [self._get_utilization(t) for t in self.device_types] or [0]

        return max(values)

    def _get_utilization(self, device_type: GpuType) -> float:
        if device_type == "amd":
            return self._get_amd_utilization()
        elif device_type == "nvidia":
            return self._get_nvidia_utilization()
        else:
            raise NotImplementedError(f"Unsupported gpu type: {device_type}")

    @staticmethod
    def _get_amd_utilization() -> float:
        import pyadl

        devices = pyadl.ADLManager.getInstance().getDevices()

        return max([d.getCurrentUsage() for d in devices])

    @staticmethod
    def _get_nvidia_utilization() -> float:
        import GPUtil

        gpus = GPUtil.getGPUs()

        return max([gpu.load * 100 for gpu in gpus])



# Python Modules
import abc
import logging
import statistics
import time
import threading

from collections import deque

# 3rd Party Modules

# Project Modules
from sleepguard.monitors import debug


log = logging.getLogger(__name__)


class UtilizationMonitor(threading.Thread, abc.ABC):
    """
    A base class for all system utilization monitoring threads.
    """
    def __init__(
            self,
            poll_period: float,
            log_history_size: int,
            threshold: float,
            *args,
            **kwargs
    ):
        """

        :param poll_period: The number of seconds between logging the system
               resource.
        :param log_history_size: The number of values to store in the
               resource utilization logs. The average utilization will
               be calculated from these values.
        :param threshold: The value at which sleep will be inhibited if the
               average utilization value exceeds it.
        """
        super().__init__(*args, **kwargs)

        if poll_period <= 0:
            raise ValueError("poll_period must be greater than zero")

        if log_history_size <= 0:
            raise ValueError("log_history_size must be greater than zero")

        if threshold < 0:
            raise ValueError("threshold must be positive")

        self._poll_period = poll_period
        self._log_history_size = log_history_size
        self._threshold = threshold
        self._logs = deque(maxlen=log_history_size)
        self._is_stopped = threading.Event()
        self._lock = threading.Lock()

    def stop(self):
        """
        Terminate the execution of this monitor.
        """
        self._is_stopped.set()

    @property
    def is_stopped(self) -> bool:
        """
        Checks if the thread is stopped.

        :return:
        """
        return self._is_stopped.is_set()

    @property
    def is_running(self) -> bool:
        """
        Checks if the thread is running (not stopped).

        :return:
        """
        return not self._is_stopped.is_set()

    @property
    def poll_period(self) -> float:
        """
        How often the value of the system resource is polled.

        :return:
        """
        return self._poll_period

    @property
    def threshold(self) -> float:
        """
        The value at which sleep will be inhibited if the average utilization
        value exceeds it.

        :return:
        """
        return self._threshold

    def run(self):
        """
        The main thread loop.

        :return:
        """
        while not self.is_stopped:
            value = self.get_instantaneous_value()
            with self._lock:
                self._logs.appendleft(value)

            log.debug(f"Sleeping for poll_period: {self._poll_period}")
            time.sleep(self._poll_period)

    @abc.abstractmethod
    def get_instantaneous_value(self) -> float:
        """
        Return the instantaneous value of the monitored system resource.

        :return:
        """
        pass

    def get_current_logs(self) -> list[float]:
        """
        Get the current logged values as a list.

        :return:
        """
        with self._lock:
            return list(self._logs)

    @debug
    def get_average_utilization(self) -> float:
        """
        Get the average utilization of the monitored resource over the
        ``threshold_period``.

        :return:
        """
        logs = self.get_current_logs()
        if not logs:
            return 0.0

        return statistics.mean(logs)

    def is_above_threshold(self) -> bool:
        """
        Returns true if the value from ``get_average_utilization`` is above the
        ``threshold``.

        :return:
        """
        return self.get_average_utilization() > self.threshold



"""
A script that monitors several types of activity and prevents the system from
sleeping if the average activity level is above a threshold over a specified
period of time.
"""

# Python Modules
from __future__ import annotations

import dataclasses
import logging
import logging.handlers
import os
import pathlib
import shutil
import signal
import sys
import time

from dataclasses import dataclass, field
from importlib.resources import as_file, files
from pprint import pformat
from typing import Any, Iterable, Optional


# 3rd Party Modules
import platformdirs

from simple_parsing import ArgumentParser, DashVariant, Serializable
from simple_parsing.utils import Dataclass
from wakepy import keep

# Project Modules
from sleepguard import resources
from sleepguard import DEFAULT_LOG_FORMAT, DEFAULT_LOG_LEVEL, DEFAULT_MONITOR_CRITERIA
from sleepguard.monitors import DEFAULT_BURN_IN, DEFAULT_POLL_PERIOD, MonitorCriteria
from sleepguard.monitors.base import UtilizationMonitor
from sleepguard.monitors.factory import UtilizationMonitorFactory


APP_NAME = "sleepguard"

system_monitors: list[UtilizationMonitor] = []


@dataclass(kw_only=True)
class ProgramOptions(Dataclass):
    """
    General options for the program.
    """

    config_path: Optional[pathlib.Path] = field(default=None)
    """
    The path to a configuration file. This will override any other
    configuration files found in supported locations.
    """

    configuration: Optional[SleepGuardConfiguration] = field(default=None)
    """
    Override individual configuration options specified in the loaded 
    configuration file.
    """


@dataclass(kw_only=True)
class SleepGuardConfiguration(Serializable):
    monitor_criteria: list[MonitorCriteria] = field(default_factory=list)

    poll_period: Optional[int] = None
    """
    The number of seconds the script will check to see if the monitored
    system resources are above their respective thresholds.
    """

    burn_in: Optional[int] = None
    """
    The monitored statistics will sensitive to outliers early in the monitoring,
    so we will not start guarding until after the burn in period to allow
    some data points to be acquired.
    """

    log_level: Optional[int] = None
    """
    The python log level to use (as an integer)
    """

    log_format: Optional[str] = None
    """
    The log format.
    """

    syslog_address: Optional[str] = None
    """
    The address of the syslog server.
    """

    def __post_init__(self):
        self.monitor_criteria = [MonitorCriteria(c) for c in self.monitor_criteria]


def merge_configurations(configurations: list[SleepGuardConfiguration]) -> SleepGuardConfiguration:
    """
    Merge multiple configuration values into a single configuration value by
    prioritizing the most local configuration values.

    :param configurations:
    :return:
    """
    final_monitor_criteria: dict[str, MonitorCriteria] = dict()
    final_fields: dict[str, Any] = dict()

    # Loop through each configuration in priority order
    for current_config in configurations:
        merge_monitor_criteria(current_config, final_monitor_criteria)
        merge_fields(current_config, final_fields)

    return SleepGuardConfiguration(
        monitor_criteria=list(final_monitor_criteria.values()),
        **final_fields
    )


def merge_monitor_criteria(
        current_config: SleepGuardConfiguration,
        final_monitor_criteria: dict[str, MonitorCriteria]
):
    """
    Update the monitor criteria from the current configuration.

    :param current_config:
    :param final_monitor_criteria:

    :return:
    """

    current_monitor_criteria = current_config.monitor_criteria

    for criteria in current_monitor_criteria:
        final_criteria = final_monitor_criteria.get(criteria.system_resource)
        if final_criteria is None:
            final_criteria = criteria
            final_monitor_criteria[criteria.system_resource] = final_criteria
        else:
            if criteria.system_resource is not None:
                final_criteria.system_resource = criteria.system_resource
            if criteria.threshold is not None:
                final_criteria.threshold = criteria.threshold
            if criteria.poll_period is not None:
                final_criteria.poll_period = criteria.poll_period
            if criteria.log_history_size is not None:
                final_criteria.log_history_size = criteria.log_history_size


def merge_fields(
        current_config: SleepGuardConfiguration,
        final_fields: dict[str, Any]
):
    """

    :param current_config:
    :param final_fields:
    :return:
    """
    # All the fields except the `monitor_criteria` which is handled separately.
    fields = [
        f
        for f in dataclasses.fields(SleepGuardConfiguration)
        if f.name != "monitor_criteria"
    ]
    for config_field in fields:
        value = getattr(current_config, config_field.name)
        if value is not None:
            final_fields[config_field.name] = value


def assign_defaults(config: SleepGuardConfiguration) -> SleepGuardConfiguration:
    """
    Assign default values for configuration values that have not been set.

    :param config:
    :return:
    """
    if config.poll_period is None:
        sys.stdout.write(f"No poll_period was specified. Using: {DEFAULT_POLL_PERIOD}\n")
        config.poll_period = DEFAULT_POLL_PERIOD
    if config.burn_in is None:
        sys.stdout.write(f"No burn_in was specified. Using: {DEFAULT_BURN_IN}\n")
        config.burn_in = DEFAULT_BURN_IN
    if config.log_level is None:
        sys.stdout.write(f"No log_level was specified. Using: {DEFAULT_LOG_LEVEL}\n")
        config.log_level = DEFAULT_LOG_LEVEL
    if config.log_format is None:
        sys.stdout.write(f"No log_format was specified. Using: {DEFAULT_LOG_FORMAT}\n")
        config.log_format = DEFAULT_LOG_FORMAT

    if not config.monitor_criteria:
        sys.stdout.write(f"No monitor_criteria were specified. Using: the defaults.\n")
        config.monitor_criteria = DEFAULT_MONITOR_CRITERIA

    return config


def load_configuration(
        config_path: Optional[pathlib.Path],
        command_line_config: Optional[SleepGuardConfiguration]
) -> SleepGuardConfiguration:
    """
    Load the project settings.

    :return:
    """
    # If a config path is given on the command line, use it.
    if config_path is not None:
        sys.stdout.write(f"Using the config file specified on the command line: {config_path}\n")
        return SleepGuardConfiguration.load(config_path)

    # Get a list of the existing configuration files.
    config_paths = get_existing_configuration_files()

    if config_paths:
        # Get the highest priority config file
        config_path = config_paths[-1]
        sys.stdout.write(f"Found config file at: {config_path}\n")
    else:
        # Install and get the default configuration file
        config_path = install_default_configuration()

    # Add the base configuration to a list
    configurations = [SleepGuardConfiguration.load(config_path)]

    # If any configuration options are added on the command line, create a
    # config object from them and add them to the list.
    if command_line_config is not None:
        sys.stdout.write(
            f"Using command line options to override values in the config file: "
            f"{command_line_config}\n"
        )
        configurations.append(command_line_config)

    # Merge the configurations by overriding the base configuration with any
    # options given on the command line.
    merged_configuration = merge_configurations(configurations)

    # If any values are still missing from the configuration, use their default
    # values.
    final_configuration = assign_defaults(merged_configuration)
    sys.stdout.write(f"Using final configuration:\n{pformat(final_configuration)}\n")

    return final_configuration


def get_existing_configuration_files() -> list[pathlib.Path]:
    """
    Find and return any existing config files in the supported system or
    user directories.

    :return:
    """
    # Get the known paths where configuration files could be stored.
    configuration_paths = get_supported_configuration_paths()

    # Only keep the files that actually exist.
    return [p for p in configuration_paths if p.is_file()]


def get_supported_configuration_paths() -> list[pathlib.Path]:
    """
    Get a list of the supported configuration paths in order from the least
    priority to the highest.

    :return:
    """
    paths = platformdirs.site_config_dir(APP_NAME, multipath=True).split(os.pathsep)
    paths += [platformdirs.user_config_dir(APP_NAME)]

    return [pathlib.Path(p) / f"{APP_NAME}.yaml" for p in paths]


def is_system_util_above_threshold(
        monitors: Iterable[UtilizationMonitor]
) -> Optional[UtilizationMonitor]:
    """
    Checks if any of the measured utilization types are above their
    corresponding threshold.

    :param monitors:
    :return: The first monitor that is above their threshold or ``None`` if all
             are below their respective thresholds.
    """

    for monitor in monitors:
        if monitor.is_above_threshold():
            return monitor

    return None


def get_system_monitors(config: SleepGuardConfiguration) -> list[UtilizationMonitor]:
    """
    Get the list of system resource monitors to use.

    :return:
    """
    monitors = [UtilizationMonitorFactory.from_criteria(c) for c in config.monitor_criteria]

    # If a monitor could not be created, e.g., an unsupported GPU, a null
    # value will be returned, and we remove it here.
    monitors = [m for m in monitors if m is not None]

    return monitors


def get_sleep_inhibitor_method() -> str:
    """
    Gets the method used to keep the system awake.

    :return:
    """
    with keep.running() as r:
        method = r.used_method
        time.sleep(0.1)

    return method


def get_default_configuration_path() -> pathlib.Path:
    """
    Get the absolute path of the default configuration file.

    :return:
    """
    with as_file(files(resources).joinpath("default_configuration.yaml")) as path:
        return path


def install_default_configuration() -> pathlib.Path:
    """
    Install a default configuration in the user's home directory, if no
    configuration is detected in any of the supported locations.

    :return: The path to the newly created default configuration file.
    """
    user_dir = pathlib.Path(platformdirs.user_config_dir(APP_NAME))
    user_config_file = user_dir / f"{APP_NAME}.yaml"
    default_config_file = get_default_configuration_path()

    user_dir.mkdir(parents=True, exist_ok=True)

    sys.stdout.write(
        f"No configuration file detected. Installing the default "
        f"configuration to: {user_config_file}\n"
    )

    shutil.copyfile(default_config_file, user_config_file)

    return user_config_file


def main(config: SleepGuardConfiguration):
    """

    :return:
    """
    # Use a global variables for these so I can stop them in the signal handler
    global system_monitors

    log = logging.getLogger(APP_NAME)
    log.info(f"Starting Sleep Guard with inhibitor: {get_sleep_inhibitor_method()}")
    log.debug(f"Using configuration: {config}")

    # Initialize all the system resource monitors
    system_monitors.extend(get_system_monitors(config))

    # Start the monitoring threads
    [monitor.start() for monitor in system_monitors]

    log.debug(f"Starting burn in period")
    time.sleep(config.burn_in)

    log.debug(f"Starting main loop")

    # The main loop
    while any([m.is_running for m in system_monitors]):
        inhibit_monitor = is_system_util_above_threshold(system_monitors)
        if inhibit_monitor is not None:
            with keep.running():
                log.debug(
                    f"Keeping the system awake because the {inhibit_monitor.__class__.__name__} "
                    f"average value {inhibit_monitor.get_average_utilization():0.2f} is "
                    f"above its threshold of {inhibit_monitor.threshold:0.2f}"
                )
                log.debug(f"Log values: {inhibit_monitor.get_current_logs()}")

                # keep.running should reset the idle timer, so there shouldn't
                # be a need to sleep for a long time.
                time.sleep(1)

        log.debug(f"Will recheck system utilization in {config.poll_period} seconds")
        time.sleep(config.poll_period)


def signal_handler(sig, frame):
    """
    Log that we are shutting down if Ctrl+C is pressed or SIGTERM is received.

    :param sig:
    :param frame:
    :return:
    """
    log = logging.getLogger(APP_NAME)
    log.info(f"Terminating Sleep Guard")
    [m.stop() for m in system_monitors]


signal.signal(signal.SIGINT, signal_handler)
signal.signal(signal.SIGTERM, signal_handler)
signal.signal(signal.SIGQUIT, signal_handler)
signal.signal(signal.SIGABRT, signal_handler)


if __name__ == "__main__":
    parser = ArgumentParser(
        description="Prevent the system from sleeping when the system is active.",
        add_option_string_dash_variants=DashVariant.UNDERSCORE_AND_DASH
    )
    parser.add_arguments(ProgramOptions, dest="options")

    program_options = parser.parse_args().options
    configuration = load_configuration(program_options.config_path, program_options.configuration)

    logging.basicConfig(
        level=configuration.log_level,
        format=configuration.log_format,
    )

    if configuration.syslog_address is not None:
        syslog_handler = logging.handlers.SysLogHandler(address=configuration.syslog_address)
        syslog_handler.ident = f"[{APP_NAME}] "
        syslog_handler.setLevel(configuration.log_level)
        logging.getLogger().addHandler(syslog_handler)

    main(configuration)



"""
Utility methods and classes for monitoring system utilization.
"""

# Python Modules
import functools
import logging
from collections import UserDict

from dataclasses import dataclass
from typing import Callable, Literal, get_args

# 3rd Party Modules
import psutil

# Project Modules

log = logging.getLogger(__name__)


SystemResource = Literal["cpu_load", "cpu_util", "gpu_util", "net_util"]
"""
The valid types of system resources to monitor.
"""


DEFAULT_POLL_PERIOD = 2
DEFAULT_BURN_IN = 60
DEFAULT_LOG_HISTORY_SiZE = 60
DEFAULT_CPU_LOAD_LOG_HISTORY_SIZE = 1
DEFAULT_CPU_LOAD_THRESHOLD = 2.0
DEFAULT_CPU_UTIL_THRESHOLD = 25
DEFAULT_GPU_UTIL_THRESHOLD = 40
DEFAULT_NET_UTIL_THRESHOLD = 8192


def debug(fn: Callable):
    """
    A decorator that logs the given member function's return value.
    '
    :param fn:
    :return:
    """
    @functools.wraps(fn)
    def _wrapper(self, *args, **kwargs):
        cls_name = self.__class__.__name__
        method_name = fn.__name__
        value = fn(self, *args, **kwargs)

        log.debug(f"{cls_name}.{method_name} -> {value}")

        return value

    return _wrapper


class MonitorCriteria(UserDict):
    """
    The criteria used for monitoring a given system resource and determining
    if it should prevent the system from sleeping.

    .. note::
       This is implemented as a dictionary because ``simple-parsing`` does
       not support lists of dataclasses.
    """
    # system_resource: SystemResource
    # """
    # The system resource to monitor.
    # """
    #
    # threshold: float
    # """
    # If the average value of this resource over the `threshold_period` is
    # above this value then the system will be inhibited from sleeping for
    # `inhibit_duration` seconds.
    # """
    #
    # poll_period: int = DEFAULT_POLL_PERIOD
    # """
    # The system resource will be monitored and logged every `poll_period`
    # seconds.
    # """
    #
    # log_history_size: float = DEFAULT_LOG_HISTORY_SiZE
    # """
    # The maximum number of values to store in the monitor logs, which will be
    # used to calculate the average resource utilization.
    # """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setdefault("poll_period", DEFAULT_POLL_PERIOD)
        self.setdefault("log_history_size", DEFAULT_LOG_HISTORY_SiZE)

        valid_keys = ("system_resource", "threshold", "poll_period", "log_history_size")
        required_keys = ("system_resource", "threshold")

        if not all(k in valid_keys for k in self.keys()):
            raise ValueError(
                f"One or more of the MonitorCriteria keys are invalid. "
                f"They must be one of {valid_keys}"
            )

        if not all(rk in self.keys() for rk in required_keys):
            raise ValueError(f"MonitorCriteria is missing a required key: {required_keys}")

        if self.system_resource not in get_args(SystemResource):
            raise ValueError(f"Invalid system resource: {self.system_resource}")

    @property
    def system_resource(self) -> SystemResource:
        return self.data.get("system_resource")

    @system_resource.setter
    def system_resource(self, value: SystemResource):
        self.data["system_resource"] = value

    @property
    def threshold(self) -> float:
        return self.data.get("threshold")

    @threshold.setter
    def threshold(self, value: float):
        self.data["threshold"] = value

    @property
    def poll_period(self) -> float:
        return self.data.get("poll_period")

    @poll_period.setter
    def poll_period(self, value: float):
        self.data["poll_period"] = value

    @property
    def log_history_size(self) -> int:
        return self.data.get("log_history_size")

    @log_history_size.setter
    def log_history_size(self, value: int):
        self.data["log_history_size"] = value

    def as_kwargs(self) -> dict[str, int | float]:
        """
        Return the criteria as ``kwargs`` suitable for constructing a
        :class:`~monitors.base.UtilizationMonitor` instance.

        :return:
        """
        return {
            "poll_period": self.poll_period,
            "log_history_size": self.log_history_size,
            "threshold": self.threshold,
        }
